@@ -15,20 +15,19 @@
 # Public License for more details.
 #
 # You should have received a copy of the GNU Lesser General Public
-# License along with this program; if not, see <http://www.gnu.org/licenses/>.
+# License along with this program; if not, see <https://www.gnu.org/licenses/>.
 
 from __future__ import absolute_import, print_function, unicode_literals
-del absolute_import, print_function, unicode_literals
 
-import pyme
-from pyme import core, constants
+import gpg
 import support
 
-support.init_gpgme(constants.PROTOCOL_OpenPGP)
-c = core.Context()
+del absolute_import, print_function, unicode_literals
 
-source = core.Data(file=support.make_filename("cipher-1.asc"))
-sink = core.Data()
+c = gpg.Context()
+
+source = gpg.Data(file=support.make_filename("cipher-1.asc"))
+sink = gpg.Data()
 
 c.op_decrypt(source, sink)
 result = c.op_decrypt_result()
@@ -38,8 +37,8 @@ assert not result.unsupported_algorithm, \
 support.print_data(sink)
 
 # Idiomatic interface.
-with pyme.Context() as c:
-    plaintext, _, _ = c.decrypt(open(support.make_filename("cipher-1.asc")))
+with gpg.Context() as c:
+    plaintext, _, _ = c.decrypt(open(support.make_filename("cipher-1.asc")), verify=False)
     assert len(plaintext) > 0
     assert plaintext.find(b'Wenn Sie dies lesen k') >= 0, \
         'Plaintext not found'
