@@ -1,3 +1,24 @@
+/* assuan-support.c - Assuan wrappers
+ * Copyright (C) 2009 g10 Code GmbH
+ *
+ * This file is part of GPGME.
+ *
+ * GPGME is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * GPGME is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
+ */
+
+
 #if HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -33,7 +54,7 @@ _gpgme_assuan_log_cb (assuan_context_t ctx, void *hook,
   if (msg == NULL)
     return 1;
 
-  _gpgme_debug (DEBUG_ASSUAN, "%s", msg);
+  _gpgme_debug (DEBUG_ASSUAN, -1, NULL, NULL, NULL, "%s", msg);
   return 0;
 }
 
@@ -97,6 +118,9 @@ my_recvmsg (assuan_context_t ctx, assuan_fd_t fd, assuan_msghdr_t msg,
 {
   (void)ctx;
 #ifdef HAVE_W32_SYSTEM
+  (void)fd;
+  (void)msg;
+  (void)flags;
   gpg_err_set_errno (ENOSYS);
   return -1;
 #else
@@ -112,6 +136,9 @@ my_sendmsg (assuan_context_t ctx, assuan_fd_t fd, const assuan_msghdr_t msg,
 {
   (void)ctx;
 #ifdef HAVE_W32_SYSTEM
+  (void)fd;
+  (void)msg;
+  (void)flags;
   gpg_err_set_errno (ENOSYS);
   return -1;
 #else
@@ -131,7 +158,7 @@ my_spawn (assuan_context_t ctx, pid_t *r_pid, const char *name,
 	  void (*atfork) (void *opaque, int reserved),
 	  void *atforkvalue, unsigned int flags)
 {
-  int err;
+  int err = 0;
   struct spawn_fd_item_s *fd_items;
   int i;
 
@@ -182,10 +209,58 @@ my_spawn (assuan_context_t ctx, pid_t *r_pid, const char *name,
   fd_items[i].fd = -1;
   fd_items[i].dup_to = -1;
 
-  err = _gpgme_io_spawn (name, (char*const*)argv,
-                         (IOSPAWN_FLAG_NOCLOSE | IOSPAWN_FLAG_DETACHED),
-			 fd_items, atfork, atforkvalue, r_pid);
-  if (! err)
+#ifdef HAVE_W32_SYSTEM
+  /* Fix up a potential logger fd so that on windows the fd
+   * translation can work through gpgme-w32spawn.
+   *
+   * We do this here as a hack because we would
+   * otherwise have to change assuan_api and the current
+   * plan in 2019 is to change away from this to gpgrt
+   * based IPC. */
+  if (argv)
+    {
+      int loc = 0;
+      while (argv[loc])
+        {
+          if (!strcmp ("--logger-fd", argv[loc]))
+            {
+              long logger_fd = -1;
+              char *tail;
+              int k = 0;
+              loc++;
+              if (!argv[loc])
+                {
+                  err = GPG_ERR_INV_ARG;
+                  break;
+                }
+              logger_fd = strtol (argv[loc], &tail, 10);
+              if (tail == argv[loc] || logger_fd <= 0)
+                {
+                  err = GPG_ERR_INV_ARG;
+                  break;
+                }
+              while (fd_items[k++].fd != -1)
+                {
+                  if (fd_items[k].fd == logger_fd)
+                    {
+                      fd_items[k].arg_loc = loc;
+                      break;
+                    }
+                }
+              break;
+            }
+          loc++;
+        }
+    }
+#endif
+
+  if (!err)
+    {
+      err = _gpgme_io_spawn (name, (char*const*)argv,
+                             (IOSPAWN_FLAG_NOCLOSE | IOSPAWN_FLAG_DETACHED),
+                             fd_items, atfork, atforkvalue, r_pid);
+    }
+  if (!err)
     {
       i = 0;
 
@@ -210,7 +285,10 @@ my_waitpid (assuan_context_t ctx, pid_t pid,
 {
   (void)ctx;
 #ifdef HAVE_W32_SYSTEM
-  CloseHandle ((HANDLE) pid);
+  (void)nowait;
+  (void)status;
+  (void)options;
+  (void)pid;  /* Just a number without a kernel object.  */
 #else
   /* We can't just release the PID, a waitpid is mandatory.  But
      NOWAIT in POSIX systems just means the caller already did the
@@ -229,6 +307,11 @@ my_socketpair (assuan_context_t ctx, int namespace, int style,
 	       int protocol, assuan_fd_t filedes[2])
 {
 #ifdef HAVE_W32_SYSTEM
+  (void)ctx;
+  (void)namespace;
+  (void)style;
+  (void)protocol;
+  (void)filedes;
   gpg_err_set_errno (ENOSYS);
   return -1;
 #else
@@ -277,4 +360,3 @@ struct assuan_system_hooks _gpgme_assuan_system_hooks =
     my_socket,
     my_connect
   };
-
