@@ -14,7 +14,8 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #if HAVE_CONFIG_H
@@ -51,6 +52,7 @@ enum
     WANT_GPG_NAME,
     WANT_GPGSM_NAME,
     WANT_G13_NAME,
+    WANT_GPG_WKS_CLIENT_NAME,
     WANT_GPG_ONE_MODE
   };
 
@@ -73,6 +75,7 @@ static struct {
   char *gpg_name;
   char *gpgsm_name;
   char *g13_name;
+  char *gpg_wks_client_name;
   int  gpg_one_mode;  /* System is in gpg1 mode.  */
 } dirinfo;
 
@@ -259,13 +262,14 @@ get_gpgconf_item (int what)
       pgmname = dirinfo.disable_gpgconf? NULL : _gpgme_get_gpgconf_path ();
       if (pgmname && access (pgmname, F_OK))
         {
-          _gpgme_debug (DEBUG_INIT,
+          _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
                         "gpgme-dinfo: gpgconf='%s' [not installed]\n", pgmname);
           free (pgmname);
           pgmname = NULL; /* Not available.  */
         }
       else
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo: gpgconf='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo: gpgconf='%s'\n",
                       pgmname? pgmname : "[null]");
       if (!pgmname)
         {
@@ -291,28 +295,36 @@ get_gpgconf_item (int what)
          allocated.  */
       dirinfo.valid = 1;
       if (dirinfo.gpg_name)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo:     gpg='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo:     gpg='%s'\n",
                       dirinfo.gpg_name);
       if (dirinfo.g13_name)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo:     g13='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo:     g13='%s'\n",
                       dirinfo.g13_name);
       if (dirinfo.gpgsm_name)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo:   gpgsm='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo:   gpgsm='%s'\n",
                       dirinfo.gpgsm_name);
       if (dirinfo.homedir)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo: homedir='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo: homedir='%s'\n",
                       dirinfo.homedir);
       if (dirinfo.agent_socket)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo:   agent='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo:   agent='%s'\n",
                       dirinfo.agent_socket);
       if (dirinfo.agent_ssh_socket)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo:     ssh='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo:     ssh='%s'\n",
                       dirinfo.agent_ssh_socket);
       if (dirinfo.dirmngr_socket)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo: dirmngr='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo: dirmngr='%s'\n",
                       dirinfo.dirmngr_socket);
       if (dirinfo.uisrv_socket)
-        _gpgme_debug (DEBUG_INIT, "gpgme-dinfo:   uisrv='%s'\n",
+        _gpgme_debug (DEBUG_INIT, -1, NULL, NULL, NULL,
+                      "gpgme-dinfo:   uisrv='%s'\n",
                       dirinfo.uisrv_socket);
     }
   switch (what)
@@ -333,6 +345,14 @@ get_gpgconf_item (int what)
     case WANT_G13_NAME:   result = dirinfo.g13_name; break;
     case WANT_UISRV_SOCKET:  result = dirinfo.uisrv_socket; break;
     case WANT_GPG_ONE_MODE: result = dirinfo.gpg_one_mode? "1":NULL; break;
+    case WANT_GPG_WKS_CLIENT_NAME:
+      if (!dirinfo.gpg_wks_client_name && dirinfo.libexecdir)
+        dirinfo.gpg_wks_client_name = _gpgme_strconcat (dirinfo.libexecdir,
+                                                        "/",
+                                                        "gpg-wks-client",
+                                                        NULL);
+      result = dirinfo.gpg_wks_client_name;
+      break;
     }
   UNLOCK (dirinfo_lock);
   return result;
@@ -438,6 +458,8 @@ gpgme_get_dirinfo (const char *what)
     return get_gpgconf_item (WANT_GPGSM_NAME);
   else if (!strcmp (what, "g13-name"))
     return get_gpgconf_item (WANT_G13_NAME);
+  else if (!strcmp (what, "gpg-wks-client-name"))
+    return get_gpgconf_item (WANT_GPG_WKS_CLIENT_NAME);
   else if (!strcmp (what, "agent-ssh-socket"))
     return get_gpgconf_item (WANT_AGENT_SSH_SOCKET);
   else if (!strcmp (what, "dirmngr-socket"))
