@@ -1,23 +1,24 @@
 /* engine-gpg.c - Gpg Engine.
-   Copyright (C) 2000 Werner Koch (dd9jn)
-   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007,
-                 2009, 2010, 2012, 2013 g10 Code GmbH
-
-   This file is part of GPGME.
-
-   GPGME is free software; you can redistribute it and/or modify it
-   under the terms of the GNU Lesser General Public License as
-   published by the Free Software Foundation; either version 2.1 of
-   the License, or (at your option) any later version.
-
-   GPGME is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with this program; if not, see <https://www.gnu.org/licenses/>.
-*/
+ * Copyright (C) 2000 Werner Koch (dd9jn)
+ * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007,
+ *               2009, 2010, 2012, 2013 g10 Code GmbH
+ *
+ * This file is part of GPGME.
+ *
+ * GPGME is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * GPGME is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
+ */
 
 #if HAVE_CONFIG_H
 #include <config.h>
@@ -141,6 +142,7 @@ struct engine_gpg
   gpgme_pinentry_mode_t pinentry_mode;
   char request_origin[10];
   char *auto_key_locate;
+  char *trust_model;
 
   struct {
     unsigned int no_symkey_cache : 1;
@@ -163,7 +165,7 @@ gpg_io_event (void *engine, gpgme_event_io_t type, void *type_data)
 {
   engine_gpg_t gpg = engine;
 
-  TRACE3 (DEBUG_ENGINE, "gpgme:gpg_io_event", gpg,
+  TRACE (DEBUG_ENGINE, "gpgme:gpg_io_event", gpg,
           "event %p, type %d, type_data %p",
           gpg->io_cbs.event, type, type_data);
   if (gpg->io_cbs.event)
@@ -455,6 +457,7 @@ gpg_release (void *engine)
   if (gpg->cmd.keyword)
     free (gpg->cmd.keyword);
   free (gpg->auto_key_locate);
+  free (gpg->trust_model);
 
   gpgme_data_release (gpg->override_session_key);
   gpgme_data_release (gpg->diagnostics);
@@ -669,6 +672,14 @@ gpg_set_engine_flags (void *engine, const gpgme_ctx_t ctx)
                                                ctx->auto_key_locate, NULL);
     }
 
+  if (ctx->trust_model && strlen (ctx->trust_model))
+    {
+      if (gpg->trust_model)
+        free (gpg->trust_model);
+      gpg->trust_model = _gpgme_strconcat ("--trust-model=",
+                                           ctx->trust_model, NULL);
+    }
+
   gpg->flags.no_symkey_cache = (ctx->no_symkey_cache
                                 && have_gpg_version (gpg, "2.2.7"));
   gpg->flags.offline = (ctx->offline && have_gpg_version (gpg, "2.1.23"));
@@ -735,7 +746,7 @@ gpg_set_status_cb (void *engine, gpgme_status_cb_t cb, void *cb_value)
 }
 
 
-/* Note, that the status_handler is allowed to modifiy the args
+/* Note, that the status_handler is allowed to modify the args
    value.  */
 static void
 gpg_set_status_handler (void *engine, engine_status_handler_t fnc,
@@ -981,6 +992,19 @@ build_argv (engine_gpg_t gpg, const char *pgmname)
       argc++;
     }
 
+  if (gpg->trust_model)
+    {
+      argv[argc] = strdup (gpg->trust_model);
+      if (!argv[argc])
+        {
+          int saved_err = gpg_error_from_syserror ();
+          free (fd_data_map);
+          free_argv (argv);
+          return saved_err;
+        }
+      argc++;
+    }
+
   if (gpg->flags.no_symkey_cache)
     {
       argv[argc] = strdup ("--no-symkey-cache");
@@ -1289,7 +1313,7 @@ read_status (engine_gpg_t gpg)
 			     received and the next thing will be that
 			     the command handler does its action.  */
 			  if (nread > 1)
-			    TRACE0 (DEBUG_CTX, "gpgme:read_status", 0,
+			    TRACE (DEBUG_CTX, "gpgme:read_status", 0,
 				    "error: unexpected data");
 
 			  add_io_cb (gpg, gpg->cmd.fd, 0,
@@ -1662,7 +1686,7 @@ gpg_decrypt (void *engine,
       if (have_gpg_version (gpg, "2.1.16"))
         {
           gpgme_data_release (gpg->override_session_key);
-          TRACE2 (DEBUG_ENGINE, "override", gpg, "seskey='%s' len=%zu\n",
+          TRACE (DEBUG_ENGINE, "override", gpg, "seskey='%s' len=%zu\n",
                   override_session_key,
                   strlen (override_session_key));
 
@@ -1671,7 +1695,14 @@ gpg_decrypt (void *engine,
                                          strlen (override_session_key), 1);
           if (!err)
             {
-              err = add_arg (gpg, "--override-session-key-fd");
+              /* We add --no-keyring because a keyring is not required
+               * when we are overriding the session key.  It would
+               * work without that option but --no-keyring avoids that
+               * gpg return a failure due to a missing key log_error()
+               * diagnostic.  --no-keyring is supported since 2.1.14. */
+              err = add_arg (gpg, "--no-keyring");
+              if (!err)
+                err = add_arg (gpg, "--override-session-key-fd");
               if (!err)
                 err = add_data (gpg, gpg->override_session_key, -2, 0);
             }
