@@ -1,21 +1,22 @@
 /* posix-io.c - Posix I/O functions
-   Copyright (C) 2000 Werner Koch (dd9jn)
-   Copyright (C) 2001, 2002, 2004, 2005, 2007, 2010 g10 Code GmbH
-
-   This file is part of GPGME.
-
-   GPGME is free software; you can redistribute it and/or modify it
-   under the terms of the GNU Lesser General Public License as
-   published by the Free Software Foundation; either version 2.1 of
-   the License, or (at your option) any later version.
-
-   GPGME is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * Copyright (C) 2000 Werner Koch (dd9jn)
+ * Copyright (C) 2001, 2002, 2004, 2005, 2007, 2010 g10 Code GmbH
+ *
+ * This file is part of GPGME.
+ *
+ * GPGME is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * GPGME is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
  */
 
 #ifdef HAVE_CONFIG_H
@@ -47,10 +48,11 @@
 #include <ctype.h>
 #include <sys/resource.h>
 
-#if __linux__
+#ifdef USE_LINUX_GETDENTS
+# include <sys/syscall.h>
 # include <sys/types.h>
 # include <dirent.h>
-#endif /*__linux__ */
+#endif /*USE_LINUX_GETDENTS*/
 
 
 #include "util.h"
@@ -59,6 +61,7 @@
 #include "ath.h"
 #include "debug.h"
 
+
 
 void
 _gpgme_io_subsystem_init (void)
@@ -106,8 +109,8 @@ int
 _gpgme_io_read (int fd, void *buffer, size_t count)
 {
   int nread;
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_read", fd,
-	      "buffer=%p, count=%u", buffer, count);
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_read", fd,
+	      "buffer=%p, count=%zu", buffer, count);
 
   do
     {
@@ -115,7 +118,7 @@ _gpgme_io_read (int fd, void *buffer, size_t count)
     }
   while (nread == -1 && errno == EINTR);
 
-  TRACE_LOGBUF (buffer, nread);
+  TRACE_LOGBUFX (buffer, nread);
   return TRACE_SYSRES (nread);
 }
 
@@ -124,9 +127,9 @@ int
 _gpgme_io_write (int fd, const void *buffer, size_t count)
 {
   int nwritten;
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_write", fd,
-	      "buffer=%p, count=%u", buffer, count);
-  TRACE_LOGBUF (buffer, count);
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_write", fd,
+	      "buffer=%p, count=%zu", buffer, count);
+  TRACE_LOGBUFX (buffer, count);
 
   do
     {
@@ -143,7 +146,7 @@ _gpgme_io_pipe (int filedes[2], int inherit_idx)
 {
   int saved_errno;
   int err;
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_pipe", filedes,
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_pipe", filedes,
 	      "inherit_idx=%i (GPGME uses it for %s)",
 	      inherit_idx, inherit_idx ? "reading" : "writing");
 
@@ -163,7 +166,8 @@ _gpgme_io_pipe (int filedes[2], int inherit_idx)
   if (err)
     return TRACE_SYSRES (err);
 
-  return TRACE_SUC2 ("read=0x%x, write=0x%x", filedes[0], filedes[1]);
+  TRACE_SUC ("read=0x%x, write=0x%x", filedes[0], filedes[1]);
+  return 0;
 }
 
 
@@ -175,7 +179,7 @@ _gpgme_io_close (int fd)
   void *handler_value;
   int idx;
 
-  TRACE_BEG (DEBUG_SYSIO, "_gpgme_io_close", fd);
+  TRACE_BEG (DEBUG_SYSIO, "_gpgme_io_close", fd, "");
 
   if (fd == -1)
     {
@@ -200,7 +204,7 @@ _gpgme_io_close (int fd)
   UNLOCK (notify_table_lock);
   if (handler)
     {
-      TRACE_LOG2 ("invoking close handler %p/%p", handler, handler_value);
+      TRACE_LOG  ("invoking close handler %p/%p", handler, handler_value);
       handler (fd, handler_value);
     }
 
@@ -217,7 +221,7 @@ _gpgme_io_set_close_notify (int fd, _gpgme_close_notify_handler_t handler,
   int res = 0;
   int idx;
 
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_set_close_notify", fd,
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_set_close_notify", fd,
 	      "close_handler=%p/%p", handler, value);
 
   assert (fd != -1);
@@ -268,7 +272,7 @@ _gpgme_io_set_nonblocking (int fd)
 {
   int flags;
   int res;
-  TRACE_BEG (DEBUG_SYSIO, "_gpgme_io_set_nonblocking", fd);
+  TRACE_BEG (DEBUG_SYSIO, "_gpgme_io_set_nonblocking", fd, "");
 
   flags = fcntl (fd, F_GETFL, 0);
   if (flags == -1)
@@ -279,6 +283,22 @@ _gpgme_io_set_nonblocking (int fd)
 }
 
 
+#ifdef USE_LINUX_GETDENTS
+/* This is not declared in public headers; getdents64(2) says that we must
+ * define it ourselves.  */
+struct linux_dirent64
+{
+  ino64_t d_ino;
+  off64_t d_off;
+  unsigned short d_reclen;
+  unsigned char d_type;
+  char d_name[];
+};
+
+# define DIR_BUF_SIZE 1024
+#endif /*USE_LINUX_GETDENTS*/
+
+
 static long int
 get_max_fds (void)
 {
@@ -287,29 +307,53 @@ get_max_fds (void)
   int rc;
 
   /* Under Linux we can figure out the highest used file descriptor by
-   * reading /proc/self/fd.  This is in the common cases much fast than
-   * for example doing 4096 close calls where almost all of them will
-   * fail.  */
-#ifdef __linux__
+   * reading /proc/self/fd.  This is in the common cases much faster
+   * than for example doing 4096 close calls where almost all of them
+   * will fail.
+   *
+   * We can't use the normal opendir/readdir/closedir interface between
+   * fork and exec in a multi-threaded process because opendir uses
+   * malloc and thus a mutex which may deadlock with a malloc in another
+   * thread.  However, the underlying getdents system call is safe.  */
+#ifdef USE_LINUX_GETDENTS
   {
-    DIR *dir = NULL;
-    struct dirent *dir_entry;
+    int dir_fd;
+    char dir_buf[DIR_BUF_SIZE];
+    struct linux_dirent64 *dir_entry;
+    int r, pos;
     const char *s;
     int x;
 
-    dir = opendir ("/proc/self/fd");
-    if (dir)
+    dir_fd = open ("/proc/self/fd", O_RDONLY | O_DIRECTORY);
+    if (dir_fd != -1)
       {
-        while ((dir_entry = readdir (dir)))
+        for (;;)
           {
-            s = dir_entry->d_name;
-            if ( *s < '0' || *s > '9')
-              continue;
-            x = atoi (s);
-            if (x > fds)
-              fds = x;
+            r = syscall(SYS_getdents64, dir_fd, dir_buf, DIR_BUF_SIZE);
+            if (r == -1)
+              {
+                /* Fall back to other methods.  */
+                fds = -1;
+                break;
+              }
+            if (r == 0)
+              break;
+
+            for (pos = 0; pos < r; pos += dir_entry->d_reclen)
+              {
+                dir_entry = (struct linux_dirent64 *) (dir_buf + pos);
+                s = dir_entry->d_name;
+                if (*s < '0' || *s > '9')
+                  continue;
+                /* atoi is not guaranteed to be async-signal-safe.  */
+                for (x = 0; *s >= '0' && *s <= '9'; s++)
+                  x = x * 10 + (*s - '0');
+                if (!*s && x > fds && x != dir_fd)
+                  fds = x;
+              }
           }
-        closedir (dir);
+
+        close (dir_fd);
       }
     if (fds != -1)
       {
@@ -317,7 +361,7 @@ get_max_fds (void)
         source = "/proc";
       }
     }
-#endif /* __linux__ */
+#endif /*USE_LINUX_GETDENTS*/
 
 #ifdef RLIMIT_NOFILE
   if (fds == -1)
@@ -384,7 +428,7 @@ get_max_fds (void)
     }
 #endif
 
-  TRACE2 (DEBUG_SYSIO, "gpgme:max_fds", 0, "max fds=%i (%s)", fds, source);
+  TRACE (DEBUG_SYSIO, "gpgme:max_fds", 0, "max fds=%ld (%s)", fds, source);
   return fds;
 }
 
@@ -430,19 +474,19 @@ _gpgme_io_spawn (const char *path, char *const argv[], unsigned int flags,
   int status;
   int signo;
 
-  TRACE_BEG1 (DEBUG_SYSIO, "_gpgme_io_spawn", path,
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_spawn", path,
 	      "path=%s", path);
   i = 0;
   while (argv[i])
     {
-      TRACE_LOG2 ("argv[%2i] = %s", i, argv[i]);
+      TRACE_LOG  ("argv[%2i] = %s", i, argv[i]);
       i++;
     }
   for (i = 0; fd_list[i].fd != -1; i++)
     if (fd_list[i].dup_to == -1)
-      TRACE_LOG2 ("fd[%i] = 0x%x", i, fd_list[i].fd);
+      TRACE_LOG  ("fd[%i] = 0x%x", i, fd_list[i].fd);
     else
-      TRACE_LOG3 ("fd[%i] = 0x%x -> 0x%x", i, fd_list[i].fd, fd_list[i].dup_to);
+      TRACE_LOG  ("fd[%i] = 0x%x -> 0x%x", i, fd_list[i].fd, fd_list[i].dup_to);
 
   pid = fork ();
   if (pid == -1)
@@ -453,10 +497,9 @@ _gpgme_io_spawn (const char *path, char *const argv[], unsigned int flags,
       /* Intermediate child to prevent zombie processes.  */
       if ((pid = fork ()) == 0)
 	{
-	  int max_fds = get_max_fds ();
-	  int fd;
-
 	  /* Child.  */
+          int max_fds = -1;
+          int fd;
 	  int seen_stdin = 0;
 	  int seen_stdout = 0;
 	  int seen_stderr = 0;
@@ -464,15 +507,40 @@ _gpgme_io_spawn (const char *path, char *const argv[], unsigned int flags,
 	  if (atfork)
 	    atfork (atforkvalue, 0);
 
-	  /* First close all fds which will not be inherited.  */
-	  for (fd = 0; fd < max_fds; fd++)
-	    {
-	      for (i = 0; fd_list[i].fd != -1; i++)
-		if (fd_list[i].fd == fd)
-		  break;
-	      if (fd_list[i].fd == -1)
-		close (fd);
-	    }
+          /* First close all fds which will not be inherited.  If we
+           * have closefrom(2) we first figure out the highest fd we
+           * do not want to close, then call closefrom, and on success
+           * use the regular code to close all fds up to the start
+           * point of closefrom.  Note that Solaris' and FreeBSD's closefrom do
+           * not return errors.  */
+#ifdef HAVE_CLOSEFROM
+          {
+            fd = -1;
+            for (i = 0; fd_list[i].fd != -1; i++)
+              if (fd_list[i].fd > fd)
+                fd = fd_list[i].fd;
+            fd++;
+#if defined(__sun) || defined(__FreeBSD__)
+            closefrom (fd);
+            max_fds = fd;
+#else /*!__sun */
+            while ((i = closefrom (fd)) && errno == EINTR)
+              ;
+            if (!i || errno == EBADF)
+              max_fds = fd;
+#endif /*!__sun*/
+          }
+#endif /*HAVE_CLOSEFROM*/
+          if (max_fds == -1)
+            max_fds = get_max_fds ();
+          for (fd = 0; fd < max_fds; fd++)
+            {
+              for (i = 0; fd_list[i].fd != -1; i++)
+                if (fd_list[i].fd == fd)
+                  break;
+              if (fd_list[i].fd == -1)
+                close (fd);
+            }
 
 	  /* And now dup and close those to be duplicated.  */
 	  for (i = 0; fd_list[i].fd != -1; i++)
@@ -501,7 +569,7 @@ _gpgme_io_spawn (const char *path, char *const argv[], unsigned int flags,
 #if 0
 		  /* FIXME: The debug file descriptor is not
 		     dup'ed anyway, so we can't see this.  */
-		  TRACE_LOG1 ("dup2 failed in child: %s\n",
+		  TRACE_LOG  ("dup2 failed in child: %s\n",
 			      strerror (errno));
 #endif
 		  _exit (8);
@@ -551,7 +619,7 @@ _gpgme_io_spawn (const char *path, char *const argv[], unsigned int flags,
 	_exit (0);
     }
 
-  TRACE_LOG1 ("waiting for child process pid=%i", pid);
+  TRACE_LOG  ("waiting for child process pid=%i", pid);
   _gpgme_io_waitpid (pid, 1, &status, &signo);
   if (status)
     return TRACE_SYSRES (-1);
@@ -586,8 +654,8 @@ _gpgme_io_select (struct io_select_fd_s *fds, size_t nfds, int nonblock)
   /* Use a 1s timeout.  */
   struct timeval timeout = { 1, 0 };
   void *dbg_help = NULL;
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_select", fds,
-	      "nfds=%u, nonblock=%u", nfds, nonblock);
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_select", fds,
+	      "nfds=%zu, nonblock=%u", nfds, nonblock);
 
   FD_ZERO (&readfds);
   FD_ZERO (&writefds);
@@ -607,7 +675,7 @@ _gpgme_io_select (struct io_select_fd_s *fds, size_t nfds, int nonblock)
           if (fds[i].fd >= FD_SETSIZE)
             {
               TRACE_END (dbg_help, " -BAD- ]");
-              gpg_err_set_errno (EBADF);
+              gpg_err_set_errno (EMFILE);
               return TRACE_SYSRES (-1);
             }
 	  assert (!FD_ISSET (fds[i].fd, &readfds));
@@ -622,7 +690,7 @@ _gpgme_io_select (struct io_select_fd_s *fds, size_t nfds, int nonblock)
           if (fds[i].fd >= FD_SETSIZE)
             {
               TRACE_END (dbg_help, " -BAD- ]");
-              gpg_err_set_errno (EBADF);
+              gpg_err_set_errno (EMFILE);
               return TRACE_SYSRES (-1);
             }
 	  assert (!FD_ISSET (fds[i].fd, &writefds));
@@ -692,7 +760,7 @@ _gpgme_io_recvmsg (int fd, struct msghdr *msg, int flags)
   int nread;
   int saved_errno;
   struct iovec *iov;
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_recvmsg", fd,
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_recvmsg", fd,
 	      "msg=%p, flags=%i", msg, flags);
 
   nread = 0;
@@ -703,7 +771,7 @@ _gpgme_io_recvmsg (int fd, struct msghdr *msg, int flags)
       iov++;
     }
 
-  TRACE_LOG1 ("about to receive %d bytes", nread);
+  TRACE_LOG  ("about to receive %d bytes", nread);
 
   do
     {
@@ -719,7 +787,7 @@ _gpgme_io_recvmsg (int fd, struct msghdr *msg, int flags)
       while (nr > 0)
 	{
 	  int len = nr > iov->iov_len ? iov->iov_len : nr;
-	  TRACE_LOGBUF (msg->msg_iov->iov_base, len);
+	  TRACE_LOGBUFX (msg->msg_iov->iov_base, len);
 	  iov++;
 	  nr -= len;
 	}
@@ -734,7 +802,7 @@ _gpgme_io_sendmsg (int fd, const struct msghdr *msg, int flags)
 {
   int nwritten;
   struct iovec *iov;
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_sendmsg", fd,
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_sendmsg", fd,
 	      "msg=%p, flags=%i", msg, flags);
 
   nwritten = 0;
@@ -745,12 +813,12 @@ _gpgme_io_sendmsg (int fd, const struct msghdr *msg, int flags)
       iov++;
     }
 
-  TRACE_LOG1 ("about to receive %d bytes", nwritten);
+  TRACE_LOG  ("about to receive %d bytes", nwritten);
   iov = msg->msg_iov;
   while (nwritten > 0)
     {
       int len = nwritten > iov->iov_len ? iov->iov_len : nwritten;
-      TRACE_LOGBUF (msg->msg_iov->iov_base, len);
+      TRACE_LOGBUFX (msg->msg_iov->iov_base, len);
       iov++;
       nwritten -= len;
     }
@@ -773,7 +841,7 @@ _gpgme_io_dup (int fd)
     new_fd = dup (fd);
   while (new_fd == -1 && errno == EINTR);
 
-  TRACE1 (DEBUG_SYSIO, "_gpgme_io_dup", fd, "new fd==%i", new_fd);
+  TRACE (DEBUG_SYSIO, "_gpgme_io_dup", fd, "new fd==%i", new_fd);
 
   return new_fd;
 }
@@ -784,7 +852,7 @@ _gpgme_io_socket (int domain, int type, int proto)
 {
   int res;
 
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_socket", domain,
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_socket", domain,
 	      "type=%i, proto=%i", type, proto);
 
   res = socket (domain, type, proto);
@@ -798,7 +866,7 @@ _gpgme_io_connect (int fd, struct sockaddr *addr, int addrlen)
 {
   int res;
 
-  TRACE_BEG2 (DEBUG_SYSIO, "_gpgme_io_connect", fd,
+  TRACE_BEG  (DEBUG_SYSIO, "_gpgme_io_connect", fd,
 	      "addr=%p, addrlen=%i", addr, addrlen);
 
   do
